-- A set of different-looking buff bars. Inspired by Elkano's Buff Bars.
-- The player can left-click on a buff to cancel it.

-- Set our default placement values. If a SavedVariable file exists, then these will be overwritten after all our code is loaded. If it doesn't, then these will work as defaults.
WhimseyDots_Placement = {x = 0, y = 600}

-- Create our context so we can do rendering.
local context = UI.CreateContext("HUD")

-- Because it's currently impossible to destroy frames, we keep a reservoir of buff bars around. Buff bars work as a single unified whole, so once they're created, we can just call their "set" function.
local activeBuffBars = {}
local zombieBuffBars = {}

local lastAttach = nil  -- Attachment point for our next buff bar.
local function StartBuffBars()  -- Hide everything we have and start adding bars one at a time.
  for _, v in pairs(activeBuffBars) do
    table.insert(zombieBuffBars, v)
    v:SetVisible(false)
    v:SetPoint("TOPCENTER", UIParent, "TOPCENTER")
  end
  
  activeBuffBars = {}
  lastAttach = nil
end

local function AddBuffBar(buff, time)
  local bar = table.remove(zombieBuffBars)
  if not bar then
    -- We don't have any bars remaining, so we create a new one.
    -- Our Bars are considered single objects that can be dealt with atomically. Each one has the functionality needed to update itself.
    bar = UI.CreateFrame("Frame", "Bar", context)

    -- Set an initial vertical pin to make some of our calculations work properly
    bar:SetPoint("TOP", UIParent, "TOP")
   
    bar.text = UI.CreateFrame("Text", "Text", bar)
    bar.timer = UI.CreateFrame("Text", "Timer", bar)
    bar.icon = UI.CreateFrame("Texture", "Icon", bar)
    
    -- Solid background - this is the actual "bar" part of it.
    bar.solid = UI.CreateFrame("Frame", "Solid", bar)
    bar.solid:SetLayer(-1)  -- Put it behind every other element.
    
    bar.text:SetText("test")
    bar.text:SetHeight(bar.text:GetFullHeight())  -- Set an initial height. We'll be overriding this, we just want to make sure it works.
    
    bar.icon:SetPoint("TOPLEFT", bar, "TOPLEFT") -- The icon is pinned to the top-left corner of the bar.
    bar.icon:SetPoint("BOTTOM", bar, "BOTTOM") -- Vertically, it always fills the entire bar.
    
    bar.text:SetPoint("TOPLEFT", bar.icon, "TOPRIGHT") -- The text is pinned to the top-right corner of the icon.
    bar:SetPoint("BOTTOM", bar.text, "BOTTOM")  -- The bar is set to always be as high as the text is.
    
    bar.timer:SetPoint("TOPRIGHT", bar, "TOPRIGHT") -- The timer is pinned to the top-right corner.
    
    bar.text:SetPoint("RIGHT", bar.timer, "LEFT") -- Make sure the text doesn't overrun the timer. We'll be changing the text's height based on the contents, but we'll leave the width calculated this way.
    
    -- Set the solid bar to fill the entire buff bar.
    bar.solid:SetPoint("TOPLEFT", bar, "TOPLEFT")
    bar.solid:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT")
    
    -- This is hardcoded, but in a full fleshed-out addon, it would be set by the user.
    -- This could be done now with slash commands, but couldn't be saved yet.
    bar:SetWidth(320)
    
    -- The function that actually sets everything.
    function bar:SetBuff(buff, time)
 --     if buff.debuff then
        -- For the sake of this addon, Debuffs are somewhat larger and red-themed.
 --       self:SetBackgroundColor(0.5, 0.3, 0.3, 0.3)
 --      self.solid:SetBackgroundColor(0.5, 0.3, 0.3)
 --       self.text:SetFontSize(20)
 --       self.timer:SetFontSize(20)
 --       self.text:SetHeight(self.text:GetFullHeight())
 --     else
        -- Buffs are smaller and blue-themed.
        self:SetBackgroundColor(0.2, 0.2, 0.6, 0.3)
        self.solid:SetBackgroundColor(0.2, 0.2, 0.6)
        self.text:SetFontSize(12)
        self.timer:SetFontSize(12)
        self.text:SetHeight(self.text:GetFullHeight())
--      end
      
      -- Re-square and set the icon
      self.icon:SetWidth(self.icon:GetHeight())
      self.icon:SetTexture("Rift", buff.icon)
      
      -- Display our stacking multiple.
      if buff.stack then
        self.text:SetText(buff.name .. " x" .. buff.stack)
      else
        self.text:SetText(buff.name)
      end
      
      if buff.duration then
        self.completion = buff.begin + buff.duration
        self.duration = buff.duration
        
        -- Display everything we might have hidden.
        self.solid:SetVisible(true)
        self.timer:SetVisible(true)
        
        self:Tick(time)
      else
        self.completion = nil
        
        -- This is a permanent buff without a timer, so don't show any of that.
        self.solid:SetVisible(false)
        self.timer:SetVisible(false)
        self.timer:SetWidth(0)
      end
      
      self.debuff = buff.debuff
      self.buffid = buff.id
    end
    
    -- This is our update function, called once every frame.
    function bar:Tick(time)
      if self.completion then
        local remaining = self.completion - time
        
        if remaining < 0 then
          self.timer:SetVisible(false)
        else
          -- Update our timer.
          self.solid:SetPoint("RIGHT", bar, remaining / self.duration, nil)
          -- Rift's UI system allows you to pin points to an arbitrary place on a frame, not just the corners and edges.
          -- These two lines are equivalent:
          -- a:SetPoint("TOPLEFT", b, "BOTTOMLEFT")
          -- a:SetPoint(0, 0, b, 0, 1)
          -- Coordinates are passed in X, Y order, with 0 being the topleft and 1 being the bottomright.
          -- We're using this feature to make Rift do the position calculation for us.
          
          -- Generate the timer text string.
          if remaining >= 3600 then
            self.timer:SetText(string.format("%d:%02d:%02d", math.floor(remaining / 3600), math.floor(remaining / 60) % 60, math.floor(remaining) % 60))
          else
            self.timer:SetText(string.format("%d:%02d", math.floor(remaining / 60), math.floor(remaining) % 60))
          end
          
          -- Update the width to avoid truncation.
          self.timer:SetWidth(self.timer:GetFullWidth())
        end
      end
    end
    
    -- Finally, if we're clicked, we want to cancel whatever buff is on us.
    function bar.Event:LeftDown()
      Command.Buff.Cancel(self.buffid)
    end
  end
  table.insert(activeBuffBars, bar)
  
  -- Show the bar and set the data.
  bar:SetVisible(true)
  bar:SetBuff(buff, time)
  
  -- Attach it to the right spot.
  if not lastAttach then
    -- This is our first bar, so we're pinning it to the top center, slightly below the top edge of the screen. Most of the screen is taken by the standard UI and this is the least cluttered place we have.
    -- The downside is that lots of buffs will run down right into the player's field of view.
    bar:SetPoint("TOPCENTER", UIParent, "TOPCENTER", WhimseyDots_Placement.x, WhimseyDots_Placement.y)
  else
    -- This isn't our first bar, so we pin it to the last bar.
    if lastAttach.debuff ~= buff.debuff then
      -- We're the first debuff, so add a gap between us and the last bar.
      bar:SetPoint("TOPCENTER", lastAttach, "BOTTOMCENTER", 0, 20)
    else
      -- Otherwise, we're flush with the last bar.
      bar:SetPoint("TOPCENTER", lastAttach, "BOTTOMCENTER")
    end
  end
  lastAttach = bar
end

local function refresh(time)
  -- Get the list of buffs from the player. Note that "player" can be replaced by "player.target" to turn this into a target's-buff-bar display.
  local bufflist = Inspect.Buff.List("player.target")
  if not bufflist then return end -- If we don't get anything, then we don't currently have information about the player. This may happen when the player is logging in or teleporting long distances.
  
  local buffdetails = Inspect.Buff.Detail("player.target", bufflist) -- Retrieve the details for all of our buffs. If there were only certain buffs we cared about, we could filter them out before asking for detail.
  local playerdetails = Inspect.Unit.Lookup("player")--Inspect.Unit.Detail("player")
  
  StartBuffBars() -- Reset the buff bars and start entering buffs.
  
  -- We want to order buffs by their time remaining, splitting apart buffs and debuffs. We do the ordering here, in standard Lua code.
  local bbars = {}
  for id, buff in pairs(buffdetails) do
	if(buff.caster == playerdetails) then
		buff.id = id  -- Make a copy of the ID, because we'll need it
		table.insert(bbars, buff)
	end
  end
  
  table.sort(bbars, function (a, b)
    if a.debuff ~= b.debuff then
      return b.debuff
    end
    
    if a.duration and b.duration then return a.remaining > b.remaining end
    if not a.duration and not b.duration then return false end
    return not a.duration
  end)
  
  -- Now that we have the ordering, we just add the bars one at a time. Done!
  for k, v in ipairs(bbars) do
    AddBuffBar(v, time)
  end
end

local resynch = false
local function tick()
  if resynch then
    -- A recalculation has been queued, so go ahead and recalculate.
    refresh(Inspect.Time.Frame())
    resynch = false
  else
    -- Just do a tick refresh
    local time = Inspect.Time.Frame()
    for _, v in ipairs(activeBuffBars) do
      v:Tick(time)
    end
  end
end

-- Events. See TrionDevelopmentTools for more detail on events.

-- A buff has changed! Rewrite the buff bars!
-- Likely, we could be more efficient by making use of this information about which buffs changed/added/removed.
-- At this current time we're not doing so. However, note that the buffs *do* provide enough information to do so.
-- Another thing we're not doing is triggering only on buffs changed on the player unit. We could be doing this as well, but, again, we're currently not.
-- As we might get *many* buff change events each frame, we're simply using a buff-change event to trigger a later refresh during the tick.
table.insert(Event.Buff.Add, {function () resynch = true end, "WhimseyDots", "refresh"})
table.insert(Event.Buff.Change, {function () resynch = true end, "WhimseyDots", "refresh"})
table.insert(Event.Buff.Remove, {function () resynch = true end, "WhimseyDots", "refresh"})

-- Our update event. This triggers every single frame.
table.insert(Event.System.Update.Begin, {tick, "WhimseyDots", "refresh"})

-- Our configuration function. Call it as "/WhimseyDots 0 600" to shift the UI element around.
local function config(commandline)
  local x, y = commandline:match("(%d+) (%d+)")
  x, y = tonumber(x), tonumber(y)
  if not x or not y then
    print("Example: /WhimseyDots 0 600")
  else
    WhimseyDots_Placement.x = x
    WhimseyDots_Placement.y = y
  end
  -- If we were being clever, and only updating when buffs changed, we'd have to update our position here.
  -- Luckily, we're not being clever.
end
table.insert(Command.Slash.Register("WhimseyDots"), {config, "WhimseyDots", "config"})
